using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Xml;
using System.Linq;
using System.Collections.Generic;
using Ato.EN.IntegrationServices.Document.DataContracts;
using System.Text.RegularExpressions;

namespace Ato.EN.IntegrationServices.CodeGenerationTPAR
{

    public partial class TPAR2021Validator
    {
        /// <summary>
        /// The max parameter name length is restricted by the SBR1 and SBR2 schemas - the lowest common denominator of SBR 1 is used as the default
        /// </summary>
        private int _maxParameterNameLength;

        /// <summary>
        /// The max parameter value length is restricted by the SBR1 and SBR2 schemas
        /// </summary>
        private int _maxParameterValueLength;

        /// <summary>
        /// The SBR1 and ebms3 schemas do not allow parameter names or values to be the empty string
        /// </summary>
        private string _emptyParameterValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="TPAR2021Validator" /> class.
        /// </summary>
        /// <param name="maxParameterNameLength">Maximum length of the parameter name.</param>
        /// <param name="maxParameterValueLength">Maximum length of the parameter value.</param>
        /// <param name="emptyParameterValue">This value will be used in place of any parameter values that result in a null or empty value.</param>
        public TPAR2021Validator(int maxParameterNameLength = 20, int maxParameterValueLength = 4096, string emptyParameterValue = "EMPTY")
        {
            _maxParameterNameLength = maxParameterNameLength;
            _maxParameterValueLength = maxParameterValueLength;
            _emptyParameterValue = emptyParameterValue;
        }

        #region Functions
        private static IEnumerable<string> Union(IEnumerable<string> list1, IEnumerable<string> list2)
        {
            IEnumerable<string> response;

            if (list1 == null && list2 == null)
            {
                response = null;
            }
            else if (list1 == null)
            {
                response = list2.Distinct();
            }
            else if (list2 == null)
            {
                response = list1.Distinct();
            }
            else
            {
                response = list1.Union(list2);
            }

            return response;
        }

        private static bool IsMatch(int? field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(Convert.ToString(field.Value), expression, options);
        }

        private static bool IsMatch(string field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(field, expression, options);
        }

        // This is just context and tuple counts where they are integer values - easier that changing the parsing logic to just return the value
        private static int Count(int count)
        {
            return count;
        }


        private static int Count<T>(IEnumerable<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static int Count<T>(ICollection<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static bool exists(bool value)
        {
            return value;
        }

        private static string GetValueOrEmpty(bool? val)
        {
            return (val.HasValue) ? val.ToString().ToLower() : string.Empty;
        }

        private static string GetValueOrEmptyDate(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-dd") : string.Empty;
        }

        private static string GetValueOrEmptyDate(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-dd");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-ddTHH:mm:ssZ");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-ddTHH:mm:ssZ") : string.Empty;
        }

        private static string GetValueOrEmpty(string val)
        {
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmpty(decimal? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        private static string GetValueOrEmpty(int? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        /// <summary>
        /// Get string value between [first] a and [last] b.
        /// </summary>
        public static string Between(string value, string a, string b)
        {
            int posA = value.IndexOf(a);
            int posB = value.LastIndexOf(b);
            if (posA == -1)
            {
                return "";
            }
            if (posB == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= posB)
            {
                return "";
            }
            return value.Substring(adjustedPosA, posB - adjustedPosA);
        }

        /// <summary>
        /// Get string value after [first] a.
        /// </summary>
        public static string Before(string value, string a)
        {
            int posA = value.IndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            return value.Substring(0, posA);
        }

        /// <summary>
        /// Get string value after [last] a.
        /// </summary>
        public static string After(string value, string a)
        {
            int posA = value.LastIndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= value.Length)
            {
                return "";
            }
            return value.Substring(adjustedPosA);
        }

        private static int Length(object field)
        {
            if (field == null)
                return 0;
            else
                return field.ToString().Trim().Length;
        }

        private static bool NotSameValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            return (from x in nodes select x).Distinct().Count() != 1;
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            nodes = nodes.Where(x => x != null).ToArray();
            return !((from x in nodes select x).Distinct().Count() == nodes.Count());

        }

        private int DuplicateValueIndex(IEnumerable<object> values)
        {
            int response = 0;
            var hashset = new HashSet<object>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private int DuplicateValueIndex<T>(IEnumerable<T?> values) where T : struct
        {
            int response = 0;
            var hashset = new HashSet<T?>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private static bool IsDate(object value)
        {
            DateTime dateValue;
            return (value != null && DateTime.TryParse(value.ToString(), out dateValue));
        }

        public DateTime AsDate(string dateAsString)
        {
            DateTime response = DateTime.MinValue;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date;
            }

            return response;
        }

        public DateTime Date(DateTime? datetime)
        {
            return datetime.GetValueOrDefault().Date;
        }

        //The logic in After function expects "---" for day and "--" for month. 
        //Since hyphen was missing the date always returned null
        public DateTime? ConvertToDate(int day, int month, int year)
        {
            return ConvertToDate(day == 0 ? null : "---" + day.ToString(), month == 0 ? null : "--" + month.ToString(), year == 0 ? null : year.ToString());
        }

        public DateTime? ConvertToDate(string day, string month, string year)
        {
            DateTime? response;
            DateTime result;

            if (year == null || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        public DateTime? ConvertToDate(string day, string month, int year)
        {
            DateTime? response;
            DateTime result;

            if (year == 0 || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year.ToString() + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        private static int Day(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Day;
            }

            return response;
        }


        private static int Day(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Day;

        }


        private static string Month(string dateAsString)
        {
            string response = null;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.ToString("MMMM");
            }
            else
            {
                return "NotAMonth";
            }

            return response;
        }

        private static string Month(DateTime? date)
        {
            if (date == null)
                return "NotAMonth";
            else
                return date.Value.ToString("MMMM");
        }


        private static int MonthAsInt(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Month;
            }

            return response;
        }

        private static int MonthAsInt(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            return date.Value.Month;
        }


        private static int Year(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Year;
            }

            return response;
        }

        private static int Year(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Year;
        }


        private static int CurrentFinancialYear()
        {
            return DateToFinancialYear(DateTime.Now, 7);
        }

        private static int FinancialYear(string dateAsString)
        {
            return DateToFinancialYear(dateAsString, 7);
        }

        private static int FinancialYear(DateTime? date)
        {
            return DateToFinancialYear(date, 7);
        }

        private static int DateToFinancialYear(string dateAsString, int startingMonth)
        {
            int response = 0;
            DateTime date;
            if (DateTime.TryParse(dateAsString, out date))
            {
                response = DateToFinancialYear(date, startingMonth);
            }

            return response;
        }
        private static int DateToFinancialYear(DateTime? date, int startingMonth)
        {
            int response;
            if (date == null)
            {
                response = 0;
            }
            else
            {
                int year = date.Value.Year;
                int month = date.Value.Month;

                if (startingMonth > month)
                    response = year;
                else
                    response = year + 1;
            }
            return response;
        }


        private static int FBTYear(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
                response = FBTYear(date);
            return response;
        }

        private static int FBTYear(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            else
            {
                if (date.Value.Month > 3)
                    return date.Value.Year + 1;
                else
                    return date.Value.Year;
            }
        }

        private static DateTime? AddMonthsToDate(DateTime? dateTime, int months)
        {
            return dateTime == null ? null : (DateTime?)dateTime.Value.AddMonths(months);
        }

        private static bool IsNumeric(object value)
        {
            decimal numbervalue;
            return (value != null && decimal.TryParse(value.ToString(), out numbervalue));
        }

        private static bool NotMonetary(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(int? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }

        private static bool NotNumeric(long? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }


        private static bool OutsideRange(decimal field, decimal expression, int range)
        {
            bool response;

            response = (field < (expression - range)) || (field > (expression + range));

            return response;
        }


        private static bool FailsUSIAlgorithm(string usi, string abn)
        {
            bool response = true;
            if (usi == null || abn == null)
            {
                response = false;
            }
            else
            {
                usi = usi.Trim();
                abn = abn.Trim();
                if (usi.Length == 14)
                {
                    int numeric;
                    if (usi.Substring(0, 11) == abn && int.TryParse(usi.Substring(11, 3), out numeric))
                        response = false;

                }
                else if (usi.Length == 9)
                {

                    if (Regex.IsMatch(usi, @"^([a-zA-Z]{3}\d{4}[a-zA-Z]{2})$"))
                        response = false;
                }
            }
            return response;
        }

        private static bool FailsTANAlgorithm(string tan)
        {
            bool response;
            decimal decimalTan;

            if (tan == null)
                return false;

            tan = tan.Trim();

            if (!decimal.TryParse(tan, out decimalTan))
                return true;

            if (tan.Length != 8)
                return true;

            decimal tanSum =
                7 * int.Parse(tan.Substring(0, 1)) +
                9 * int.Parse(tan.Substring(1, 1)) +
                8 * int.Parse(tan.Substring(2, 1)) +
                4 * int.Parse(tan.Substring(3, 1)) +
                6 * int.Parse(tan.Substring(4, 1)) +
                3 * int.Parse(tan.Substring(5, 1)) +
                5 * int.Parse(tan.Substring(6, 1)) +
                1 * int.Parse(tan.Substring(7, 1));

            if ((tanSum % 11) == 0)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsABNAlgorithm(string abn)
        {
            bool response;
            decimal decimalAbn;

            if (abn == null)
                return false;

            abn = abn.Trim();

            if (!decimal.TryParse(abn, out decimalAbn))
                return true;

            if (abn.Length != 11)
                return true;

            decimal abnSum =
                10 * (int.Parse(abn.Substring(0, 1)) - 1) +
                1 * int.Parse(abn.Substring(1, 1)) +
                3 * int.Parse(abn.Substring(2, 1)) +
                5 * int.Parse(abn.Substring(3, 1)) +
                7 * int.Parse(abn.Substring(4, 1)) +
                9 * int.Parse(abn.Substring(5, 1)) +
                11 * int.Parse(abn.Substring(6, 1)) +
                13 * int.Parse(abn.Substring(7, 1)) +
                15 * int.Parse(abn.Substring(8, 1)) +
                17 * int.Parse(abn.Substring(9, 1)) +
                19 * int.Parse(abn.Substring(10, 1));

            if ((abnSum % 89) == 0)
                response = false;
            else
                response = true;

            return response;
        }
        private static bool FailsACNAlgorithm(string acn)
        {
            bool response;
            decimal decimalAbn;
            if (acn == null)
                return false;

            acn = acn.Trim();

            if (!decimal.TryParse(acn, out decimalAbn))
                return true;

            if (acn.Length != 9)
                return true;

            decimal abnSum =
                8 * int.Parse(acn.Substring(0, 1)) +
                7 * int.Parse(acn.Substring(1, 1)) +
                6 * int.Parse(acn.Substring(2, 1)) +
                5 * int.Parse(acn.Substring(3, 1)) +
                4 * int.Parse(acn.Substring(4, 1)) +
                3 * int.Parse(acn.Substring(5, 1)) +
                2 * int.Parse(acn.Substring(6, 1)) +
                1 * int.Parse(acn.Substring(7, 1));

            decimal checkDigit = int.Parse(acn.Substring(8, 1));
            decimal acnRemainder = abnSum % 10;

            if (((10 - acnRemainder) % 10) == checkDigit)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsTFNAlgorithm(string tfn)
        {
            bool response;
            decimal decimalTfn;

            if (tfn == null)
                return false;

            tfn = tfn.Trim();
            tfn = Regex.Replace(tfn, "^0+", "");

            if (!decimal.TryParse(tfn, out decimalTfn))
                return true;

            if (tfn.Length < 8)
                return true;


            decimal tfn1To7Sum =
                1 * int.Parse(tfn.Substring(0, 1)) +
                4 * int.Parse(tfn.Substring(1, 1)) +
                3 * int.Parse(tfn.Substring(2, 1)) +
                7 * int.Parse(tfn.Substring(3, 1)) +
                5 * int.Parse(tfn.Substring(4, 1)) +
                8 * int.Parse(tfn.Substring(5, 1)) +
                6 * int.Parse(tfn.Substring(6, 1));

            decimal tfn8 = 9 * int.Parse(tfn.Substring(7, 1));

            if (tfn.Length == 8)
            {
                decimal tFNLg8WSum9 = 10 * int.Parse(tfn.Substring(7, 1));
                decimal tFNLg8WSum = tfn1To7Sum + tFNLg8WSum9;

                if ((tFNLg8WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else if (tfn.Length == 9)
            {
                decimal tfn9 = 10 * int.Parse(tfn.Substring(8, 1));
                decimal tFNLg9WSum = tfn1To7Sum + tfn8 + tfn9;

                if ((tFNLg9WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else
            {
                response = true;
            }

            return response;
        }


        private static decimal ConditionalValue(bool expression, decimal? trueVal, decimal? falseVal)
        {
            return expression ? trueVal.GetValueOrDefault() : falseVal.GetValueOrDefault();
        }

        private static decimal AsNumeric(string value)
        {
            decimal numberValue;
            decimal.TryParse(value, out numberValue);
            return numberValue;
        }

        private static bool RegexMatch(int? field, string expression, string flags = "")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static bool RegexMatch(string field, string expression, string flags = "")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static RegexOptions GetRegexOption(string flags)
        {
            RegexOptions options = RegexOptions.None;

            char[] characters = flags.ToCharArray();

            foreach (char character in characters)
            {
                switch (character)
                {
                    case 'i':
                        options = options | RegexOptions.IgnoreCase;
                        break;
                    case 'm':
                        options = options | RegexOptions.Multiline;
                        break;
                    case 's':
                        options = options | RegexOptions.Singleline;
                        break;
                    case 'n':
                        options = options | RegexOptions.ExplicitCapture;
                        break;
                    case 'x':
                        options = options | RegexOptions.IgnorePatternWhitespace;
                        break;
                }
            }

            return options;
        }

        private static string Substring(object field, int start, int length)
        {
            return field.ToString().Substring(start, length);
        }

        /// <summary>
        /// Returns an occurrence index as [occurrenceIndex] of occurrenceIndex > 0, otherwise the empty string
        /// </summary>
        /// <param name="occurrenceIndex">Index of the occurrence.</param>
        /// <returns>Occurrence in XPath [#] format</returns>
        public string OccurrenceIndex(int occurrenceIndex)
        {
            return occurrenceIndex > 0 ? "[" + occurrenceIndex + "]" : "";
        }

        #endregion // Functions

        public TPAR2021 ConsumedReport { get; private set; }

        private static Dictionary<string, ProcessMessageDocument> _processMessageDocuments = new Dictionary<string, ProcessMessageDocument>();

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public List<ProcessMessageDocument> ValidateReport(TPAR2021 report)
        {

            List<ProcessMessageDocument> response = new List<ProcessMessageDocument>();
            ProcessMessageDocument processMessage;
            ProcessMessageParameter parameter;
            bool assertion;

            this.ConsumedReport = report;

            // ------------------------------------------------------------------------------
            // Validations are now done in code.
            // This version supports full validation of a report - for those rules that have successfully parsed and could be generated.
            // Generated Validations include:
            //   - Validation logic in C#.
            //   - Production of errors in EBMS, SBR and SWS format.
            // 
            // The generation at this stage does not support the following - and has to be completed manually (later versions of the generation will do this for you)
            //   - Business Rules that did not parse.  The ESR team will help support you where this happens as an effort is being made to rectify these
            //   - You will get TODO tasks for those that the parser could not cope with
            // ------------------------------------------------------------------------------


            #region VR.ATO.TPAR.000006

            /*  VR.ATO.TPAR.000006
            Period start date and end date must be within the same financial year
    
            Legacy Rule Format:
            ((FinancialYear(^TPAR153) <> ^TPAR47) OR (FinancialYear(^TPAR154) <> ^TPAR47))

            Technical Business Rule Format:
            ((FinancialYear(^TPAR153) <> ^TPAR47) OR (FinancialYear(^TPAR154) <> ^TPAR47))
    
            Data Elements:
    
            ^TPAR47 = TPAR:RP:Report.TargetFinancial.Year
    
            ^TPAR153 = TPAR:RP:Period.Start.Date
    
            ^TPAR154 = TPAR:RP:Period.End.Date
            */
            assertion = (FinancialYear(report.TPAR153.GetValueOrDefault()) != report.TPAR47.GetValueOrDefault() || FinancialYear(report.TPAR154.GetValueOrDefault()) != report.TPAR47.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000006",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Reporting party period start date and end date must be within the same financial year",
                    Location = "/tns:TPAR/tns:RP/tns:TargetFinancialY",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000006" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR153", Value = GetValueOrEmptyDate(report.TPAR153) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR47", Value = GetValueOrEmpty(report.TPAR47) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR154", Value = GetValueOrEmptyDate(report.TPAR154) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000006

            #region VR.ATO.TPAR.000011

            /*  VR.ATO.TPAR.000011
            Organisation name must contain at least one alpha character
    
            Legacy Rule Format:
            (^TPAR4 <> NULLORBLANK AND NotContainsSet(^TPAR4, '"a-z", "A-Z"'))

            Technical Business Rule Format:
            (^TPAR4 <> BLANK AND NotContainsSet(^TPAR4, '"a-z", "A-Z"'))
    
            Data Elements:
    
            ^TPAR4 = TPAR:INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR4) != true && !(IsMatch(report.TPAR4, @"[a-z]|[A-Z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000011",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name must contain at least one alpha character",
                    Location = "/tns:TPAR/tns:INT/tns:OrganisationNameDetailsOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000011" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR4", Value = report.TPAR4 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000011

            #region VR.ATO.TPAR.000012

            /*  VR.ATO.TPAR.000012
            Organisation name cannot contain repeated hyphen, apostrophe, or space characters
    
            Legacy Rule Format:
            (ContainsSet(^TPAR4, '"--","''","  "'))

            Technical Business Rule Format:
            (ContainsSet(^TPAR4, '"--","''","  "'))
    
            Data Elements:
    
            ^TPAR4 = TPAR:INT:OrganisationNameDetails:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = IsMatch(report.TPAR4, @"--|''|  ", RegexOptions.IgnoreCase);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000012",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Organisation name cannot contain repeated hyphen, apostrophe, or space characters",
                    Location = "/tns:TPAR/tns:INT/tns:OrganisationNameDetailsOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000012" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR4", Value = report.TPAR4 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000012

            #region VR.ATO.TPAR.000024

            /*  VR.ATO.TPAR.000024
            Country code must be present for overseas address
    
            Legacy Rule Format:
            (^TPAR18 = TRUE AND ^TPAR29 = NULL)

            Technical Business Rule Format:
            ^TPAR18 = TRUE AND ^TPAR29 = NULL
    
            Data Elements:
    
            ^TPAR29 = TPAR:INT:AddressDetailsBusiness:AddressDetails.Country.Code
    
            ^TPAR18 = TPAR:INT:AddressDetailsBusiness:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (report.TPAR18 == true && report.TPAR29 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000024",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Country code must be present for overseas address",
                    Location = "/tns:TPAR/tns:INT/tns:AddressDetailsBusiness/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000024" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR18", Value = GetValueOrEmpty(report.TPAR18) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR29", Value = GetValueOrEmpty(report.TPAR29) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000024

            #region VR.ATO.TPAR.000025

            /*  VR.ATO.TPAR.000025
            Email address must not start or end with an @ symbol
    
            Legacy Rule Format:
            (StartsWith(^TPAR45, '@') OR EndsWith(^TPAR45, '@'))

            Technical Business Rule Format:
            (StartsWith(^TPAR45, '@') OR EndsWith(^TPAR45, '@'))
    
            Data Elements:
    
            ^TPAR45 = TPAR:INT:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (IsMatch(report.TPAR45, @"^(@)", RegexOptions.IgnoreCase) || IsMatch(report.TPAR45, @"(@)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000025",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must not start or end with an @ symbol",
                    Location = "/tns:TPAR/tns:INT/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000025" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR45", Value = report.TPAR45 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000025

            #region VR.ATO.TPAR.000026

            /*  VR.ATO.TPAR.000026
            Email address must contain an @ symbol
    
            Legacy Rule Format:
            (^TPAR45 <> NULLORBLANK AND NotContains(^TPAR45, '@'))

            Technical Business Rule Format:
            (^TPAR45 <> BLANK AND NotContains(^TPAR45, '@'))
    
            Data Elements:
    
            ^TPAR45 = TPAR:INT:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR45) != true && !(report.TPAR45 == null ? true : report.TPAR45.ToUpperInvariant().Contains(@"@")));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000026",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must contain an @ symbol",
                    Location = "/tns:TPAR/tns:INT/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000026" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR45", Value = report.TPAR45 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000026

            #region VR.ATO.TPAR.000027

            /*  VR.ATO.TPAR.000027
            Email address must contain a dot (.)
    
            Legacy Rule Format:
            (^TPAR45 <> NULLORBLANK AND NotContains(^TPAR45, '.'))

            Technical Business Rule Format:
            (^TPAR45 <> BLANK AND NotContains(^TPAR45, '.'))
    
            Data Elements:
    
            ^TPAR45 = TPAR:INT:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR45) != true && !(report.TPAR45 == null ? true : report.TPAR45.ToUpperInvariant().Contains(@".")));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000027",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Email address must contain a dot (.)",
                    Location = "/tns:TPAR/tns:INT/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000027" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR45", Value = report.TPAR45 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000027

            #region VR.ATO.TPAR.000029

            /*  VR.ATO.TPAR.000029
            Financial Year must not be less than 2013
    
            Legacy Rule Format:
            (^TPAR47 <> NULL AND ^TPAR47 < 2013)

            Technical Business Rule Format:
            ^TPAR47 <> NULL AND ^TPAR47 < 2013
    
            Data Elements:
    
            ^TPAR47 = TPAR:RP:Report.TargetFinancial.Year
            */
            assertion = (report.TPAR47 != null && report.TPAR47.GetValueOrDefault() < 2013);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000029",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Financial Year must not be less than 2013",
                    Location = "/tns:TPAR/tns:RP/tns:TargetFinancialY",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000029" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR47", Value = GetValueOrEmpty(report.TPAR47) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000029

            #region VR.ATO.TPAR.000044

            /*  VR.ATO.TPAR.000044
            Financial year must not be greater than current financial year
    
            Legacy Rule Format:
            (^TPAR47 <> NULL AND ^TPAR47 > CurrentFinancialYear())

            Technical Business Rule Format:
            (^TPAR47 <> NULL AND ^TPAR47 > CurrentFinancialYear())
    
            Data Elements:
    
            ^TPAR47 = TPAR:RP:Report.TargetFinancial.Year
            */
            assertion = (report.TPAR47 != null && report.TPAR47.GetValueOrDefault() > CurrentFinancialYear());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000044",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Future lodgements are not supported",
                    Location = "/tns:TPAR/tns:RP/tns:TargetFinancialY",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000044" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR47", Value = GetValueOrEmpty(report.TPAR47) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000044

            #region VR.ATO.TPAR.000045

            /*  VR.ATO.TPAR.000045
            Tax Agent Number or Registered Agent Number is not valid
    
            Legacy Rule Format:
            (^TPAR131 <> NULL AND FailsTANAlgorithm (^TPAR131))

            Technical Business Rule Format:
            (^TPAR131 <> NULL AND FailsTANAlgorithm (^TPAR131))
    
            Data Elements:
    
            ^TPAR131 = TPAR:INT:Identifiers.TaxAgentNumber.Identifier
            */
            assertion = (report.TPAR131 != null && FailsTANAlgorithm(report.TPAR131));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000045",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Tax Agent Number or Registered Agent Number has failed the algorithm check",
                    Location = "/tns:TPAR/tns:INT/tns:TaxAgentNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000045" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR131", Value = GetValueOrEmpty(report.TPAR131) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000045

            #region VR.ATO.TPAR.000060

            /*  VR.ATO.TPAR.000060
            Supplier ABN must be a valid ABN
    
            Legacy Rule Format:
            FailsABNAlgorithm(^TPAR132)

            Technical Business Rule Format:
            FailsABNAlgorithm(^TPAR132)
    
            Data Elements:
    
            ^TPAR132 = TPAR:INT:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = FailsABNAlgorithm(report.TPAR132);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000060",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Supplier ABN is invalid.",
                    Location = "/tns:TPAR/tns:INT/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000060" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR132", Value = report.TPAR132 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000060

            #region VR.ATO.TPAR.000061

            /*  VR.ATO.TPAR.000061
            Payer ABN must be a valid ABN
    
            Legacy Rule Format:
            FailsABNAlgorithm(^TPAR133)

            Technical Business Rule Format:
            FailsABNAlgorithm(^TPAR133)
    
            Data Elements:
    
            ^TPAR133 = TPAR:RP:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = FailsABNAlgorithm(report.TPAR133);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000061",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"Payer ABN is invalid.",
                    Location = "/tns:TPAR/tns:RP/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000061" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR133", Value = report.TPAR133 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000061

            #region VR.ATO.TPAR.000063

            /*  VR.ATO.TPAR.000063
            Street address state and post code must be present for an Australian address
    
            Legacy Rule Format:
            (^TPAR18 = FALSE AND (^TPAR26 = NULL OR ^TPAR27 = NULL))

            Technical Business Rule Format:
            (^TPAR18 = FALSE AND (^TPAR26 = NULL OR ^TPAR27 = NULL))
    
            Data Elements:
    
            ^TPAR18 = TPAR:INT:AddressDetailsBusiness:AddressDetails.OverseasAddress.Indicator
    
            ^TPAR26 = TPAR:INT:AddressDetailsBusiness:AddressDetails.Postcode.Text
    
            ^TPAR27 = TPAR:INT:AddressDetailsBusiness:AddressDetails.StateOrTerritory.Code
            */
            assertion = (report.TPAR18 == false && (report.TPAR26 == null || report.TPAR27 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.000063",
                    Severity = ProcessMessageSeverity.Error,
                    Description = @"State and post code must be present for an Australian address",
                    Location = "/tns:TPAR/tns:INT/tns:AddressDetailsBusiness/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.000063" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR18", Value = GetValueOrEmpty(report.TPAR18) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR26", Value = GetValueOrEmpty(report.TPAR26) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR27", Value = GetValueOrEmpty(report.TPAR27) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.000063

            #region VR.ATO.TPAR.W00001

            /*  VR.ATO.TPAR.W00001
            Overseas address indicator must not be 'TRUE' if Country Code is 'au'
    
            Legacy Rule Format:
            (^TPAR68 = "au" AND ^TPAR57 = TRUE)

            Technical Business Rule Format:
            ^TPAR68 = 'au' AND ^TPAR57 = TRUE
    
            Data Elements:
    
            ^TPAR57 = TPAR:RP:AddressDetails:AddressDetails.OverseasAddress.Indicator
    
            ^TPAR68 = TPAR:RP:AddressDetails:AddressDetails.Country.Code
            */
            assertion = (report.TPAR68 == @"au" && report.TPAR57 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00001",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Overseas address indicator must not be 'TRUE' if Country Code is 'au'",
                    Location = "/tns:TPAR/tns:RP/tns:AddressDetails/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00001" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR68", Value = GetValueOrEmpty(report.TPAR68) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR57", Value = GetValueOrEmpty(report.TPAR57) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00001

            #region VR.ATO.TPAR.W00001

            /*  VR.ATO.TPAR.W00001
            Overseas address indicator must not be 'TRUE' if Country Code is 'au'
    
            Legacy Rule Format:
            (^TPAR29 = "au" AND  ^TPAR18 = TRUE)

            Technical Business Rule Format:
            ^TPAR29 = 'au' AND  ^TPAR18 = TRUE
    
            Data Elements:
    
            ^TPAR18 = TPAR:INT:AddressDetailsBusiness:AddressDetails.OverseasAddress.Indicator
    
            ^TPAR29 = TPAR:INT:AddressDetailsBusiness:AddressDetails.Country.Code
            */
            assertion = (report.TPAR29 == @"au" && report.TPAR18 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00001",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Overseas address indicator must not be 'TRUE' if Country Code is 'au'",
                    Location = "/tns:TPAR/tns:INT/tns:AddressDetailsBusiness/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00001" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR29", Value = GetValueOrEmpty(report.TPAR29) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR18", Value = GetValueOrEmpty(report.TPAR18) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00001

            #region VR.ATO.TPAR.W00005

            /*  VR.ATO.TPAR.W00005
            If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'
    
            Legacy Rule Format:
            (^TPAR68 <> NULLORBLANK AND ^TPAR68 <> "au" AND ^TPAR57 = FALSE)

            Technical Business Rule Format:
            ^TPAR68 <> BLANK AND ^TPAR68 <> 'au' AND ^TPAR57 = FALSE
    
            Data Elements:
    
            ^TPAR68 = TPAR:RP:AddressDetails:AddressDetails.Country.Code
    
            ^TPAR57 = TPAR:RP:AddressDetails:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR68) != true && report.TPAR68 != @"au" && report.TPAR57 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00005",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'",
                    Location = "/tns:TPAR/tns:RP/tns:AddressDetails/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00005" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR68", Value = GetValueOrEmpty(report.TPAR68) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR57", Value = GetValueOrEmpty(report.TPAR57) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00005

            #region VR.ATO.TPAR.W00005

            /*  VR.ATO.TPAR.W00005
            If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'
    
            Legacy Rule Format:
            (^TPAR29 <> NULLORBLANK AND ^TPAR29 <> "au" AND ^TPAR18 = FALSE)

            Technical Business Rule Format:
            ^TPAR29 <> BLANK AND ^TPAR29 <> 'au' AND ^TPAR18 = FALSE
    
            Data Elements:
    
            ^TPAR29 = TPAR:INT:AddressDetailsBusiness:AddressDetails.Country.Code
    
            ^TPAR18 = TPAR:INT:AddressDetailsBusiness:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR29) != true && report.TPAR29 != @"au" && report.TPAR18 == false);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00005",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'",
                    Location = "/tns:TPAR/tns:INT/tns:AddressDetailsBusiness/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00005" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR29", Value = GetValueOrEmpty(report.TPAR29) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR18", Value = GetValueOrEmpty(report.TPAR18) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00005

            #region VR.ATO.TPAR.W00006

            /*  VR.ATO.TPAR.W00006
            Organisation name must contain at least one alpha character
    
            Legacy Rule Format:
            (^TPAR51 <> NULLORBLANK AND NotContainsSet(^TPAR51, '"a-z", "A-Z"'))

            Technical Business Rule Format:
            (^TPAR51 <> BLANK AND NotContainsSet(^TPAR51, '"a-z", "A-Z"'))
    
            Data Elements:
    
            ^TPAR51 = TPAR:RP:OrganisationNameDetailsMain:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR51) != true && !(IsMatch(report.TPAR51, @"[a-z]|[A-Z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00006",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Organisation name must contain at least one alpha character",
                    Location = "/tns:TPAR/tns:RP/tns:OrganisationNameDetailsMainOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00006" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR51", Value = report.TPAR51 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00006

            #region VR.ATO.TPAR.W00007

            /*  VR.ATO.TPAR.W00007
            Organisation name must contain at least one alpha character
    
            Legacy Rule Format:
            (^TPAR55 <> NULLORBLANK AND NotContainsSet(^TPAR55, '"a-z", "A-Z"'))

            Technical Business Rule Format:
            (^TPAR55 <> BLANK AND NotContainsSet(^TPAR55, '"a-z", "A-Z"'))
    
            Data Elements:
    
            ^TPAR55 = TPAR:RP:OrganisationNameDetailsTrading:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR55) != true && !(IsMatch(report.TPAR55, @"[a-z]|[A-Z]", RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00006",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Organisation name must contain at least one alpha character",
                    Location = "/tns:TPAR/tns:RP/tns:OrganisationNameDetailsTradingOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00007" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR55", Value = report.TPAR55 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00007

            #region VR.ATO.TPAR.W00008

            /*  VR.ATO.TPAR.W00008
            Street address state and post code not supplied for an Australian address
    
            Legacy Rule Format:
            (^TPAR57 = FALSE AND (^TPAR65 = NULL OR ^TPAR66 = NULL))

            Technical Business Rule Format:
            (^TPAR57 = FALSE AND (^TPAR65 = NULL OR ^TPAR66 = NULL))
    
            Data Elements:
    
            ^TPAR57 = TPAR:RP:AddressDetails:AddressDetails.OverseasAddress.Indicator
    
            ^TPAR65 = TPAR:RP:AddressDetails:AddressDetails.Postcode.Text
    
            ^TPAR66 = TPAR:RP:AddressDetails:AddressDetails.StateOrTerritory.Code
            */
            assertion = (report.TPAR57 == false && (report.TPAR65 == null || report.TPAR66 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00008",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"State and post code not supplied for an Australian address",
                    Location = "/tns:TPAR/tns:RP/tns:AddressDetails/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00008" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR57", Value = GetValueOrEmpty(report.TPAR57) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR65", Value = GetValueOrEmpty(report.TPAR65) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR66", Value = GetValueOrEmpty(report.TPAR66) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00008

            #region VR.ATO.TPAR.W00011

            /*  VR.ATO.TPAR.W00011
            Country code must be present for overseas address
    
            Legacy Rule Format:
            (^TPAR57 = TRUE AND ^TPAR68 = NULL)

            Technical Business Rule Format:
            ^TPAR57 = TRUE AND ^TPAR68 = NULL
    
            Data Elements:
    
            ^TPAR68 = TPAR:RP:AddressDetails:AddressDetails.Country.Code
    
            ^TPAR57 = TPAR:RP:AddressDetails:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (report.TPAR57 == true && report.TPAR68 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00011",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Country code must be present for overseas address",
                    Location = "/tns:TPAR/tns:RP/tns:AddressDetails/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00011" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR57", Value = GetValueOrEmpty(report.TPAR57) });

                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR68", Value = GetValueOrEmpty(report.TPAR68) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00011

            #region VR.ATO.TPAR.W00013

            /*  VR.ATO.TPAR.W00013
            Email address must not start or end with an @ symbol
    
            Legacy Rule Format:
            (StartsWith(^TPAR83, '@') OR EndsWith(^TPAR83, '@'))

            Technical Business Rule Format:
            (StartsWith(^TPAR83, '@') OR EndsWith(^TPAR83, '@'))
    
            Data Elements:
    
            ^TPAR83 = TPAR:RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (IsMatch(report.TPAR83, @"^(@)", RegexOptions.IgnoreCase) || IsMatch(report.TPAR83, @"(@)$", RegexOptions.IgnoreCase));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00013",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Email address must not start or end with an @ symbol",
                    Location = "/tns:TPAR/tns:RP/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00013" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR83", Value = report.TPAR83 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00013

            #region VR.ATO.TPAR.W00014

            /*  VR.ATO.TPAR.W00014
            Email address must contain an @ symbol
    
            Legacy Rule Format:
            (^TPAR83 <> NULLORBLANK AND NotContains(^TPAR83, '@'))

            Technical Business Rule Format:
            (^TPAR83 <> BLANK AND NotContains(^TPAR83, '@'))
    
            Data Elements:
    
            ^TPAR83 = TPAR:RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR83) != true && !(report.TPAR83 == null ? true : report.TPAR83.ToUpperInvariant().Contains(@"@")));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00014",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Email address must contain an @ symbol",
                    Location = "/tns:TPAR/tns:RP/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00014" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR83", Value = report.TPAR83 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00014

            #region VR.ATO.TPAR.W00015

            /*  VR.ATO.TPAR.W00015
            Email address must contain a dot (.)
    
            Legacy Rule Format:
            (^TPAR83 <> NULLORBLANK AND NotContains(^TPAR83, '.'))

            Technical Business Rule Format:
            (^TPAR83 <> BLANK AND NotContains(^TPAR83, '.'))
    
            Data Elements:
    
            ^TPAR83 = TPAR:RP:ElectronicContactElectronicMail:ElectronicContact.ElectronicMail.Address.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPAR83) != true && !(report.TPAR83 == null ? true : report.TPAR83.ToUpperInvariant().Contains(@".")));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00015",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Email address must contain a dot (.)",
                    Location = "/tns:TPAR/tns:RP/tns:ElectronicContactElectronicMailAddressT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00015" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR83", Value = report.TPAR83 });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00015

            #region VR.ATO.TPAR.W00033

            /*  VR.ATO.TPAR.W00033
            Signature date must not be later than today
    
            Legacy Rule Format:
            (^TPAR129 <> NULL AND ^TPAR129 > Today())

            Technical Business Rule Format:
            (^TPAR129 <> NULL AND ^TPAR129 > Today())
    
            Data Elements:
    
            ^TPAR129 = TPAR:INT:Declaration:Declaration.Signature.Date
            */
            assertion = (report.TPAR129 != null && report.TPAR129.GetValueOrDefault() > DateTime.Now.Date);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00033",
                    Severity = ProcessMessageSeverity.Warning,
                    Description = @"Signature date must not be later than today",
                    Location = "/tns:TPAR/tns:INT/tns:Declaration/tns:SignatureD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() { Name = "RuleIdentifier", Value = "VR.ATO.TPAR.W00033" } },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                { Name = "TPAR129", Value = GetValueOrEmptyDate(report.TPAR129) });

                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPAR.W00033

            foreach (ProcessMessageDocument currentProcessMessage in response)
            {
                if (currentProcessMessage.Parameters != null)
                {
                    foreach (ProcessMessageParameter currentParameter in currentProcessMessage.Parameters)
                    {
                        if (string.IsNullOrEmpty(currentParameter.Name))
                        {
                            currentParameter.Name = _emptyParameterValue;
                        }

                        if (currentParameter.Name.Length > _maxParameterNameLength)
                        {
                            currentParameter.Name = currentParameter.Name.Substring(0, _maxParameterNameLength - 1);
                        }

                        if (string.IsNullOrEmpty(currentParameter.Value))
                        {
                            currentParameter.Value = _emptyParameterValue;
                        }

                        if (currentParameter.Value.Length > _maxParameterValueLength)
                        {
                            currentParameter.Value = currentParameter.Value.Substring(0, _maxParameterValueLength - 1);
                        }
                    }
                }
            }

            return response;
        }


    }
}
