using System;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Xml;
using System.Linq;
using System.Collections.Generic;
using Ato.EN.IntegrationServices.Document.DataContracts;
using System.Text.RegularExpressions;

namespace Ato.EN.IntegrationServices.CodeGenerationTPARPYE
{

    public partial class TPARPYE2021Validator
    {
        /// <summary>
        /// The max parameter name length is restricted by the SBR1 and SBR2 schemas - the lowest common denominator of SBR 1 is used as the default
        /// </summary>
        private int _maxParameterNameLength;

        /// <summary>
        /// The max parameter value length is restricted by the SBR1 and SBR2 schemas
        /// </summary>
        private int _maxParameterValueLength;

        /// <summary>
        /// The SBR1 and ebms3 schemas do not allow parameter names or values to be the empty string
        /// </summary>
        private string _emptyParameterValue;

        /// <summary>
        /// Initializes a new instance of the <see cref="TPARPYE2021Validator" /> class.
        /// </summary>
        /// <param name="maxParameterNameLength">Maximum length of the parameter name.</param>
        /// <param name="maxParameterValueLength">Maximum length of the parameter value.</param>
        /// <param name="emptyParameterValue">This value will be used in place of any parameter values that result in a null or empty value.</param>
        public TPARPYE2021Validator(int maxParameterNameLength = 20, int maxParameterValueLength = 4096, string emptyParameterValue = "EMPTY")
        {
            _maxParameterNameLength = maxParameterNameLength;
            _maxParameterValueLength = maxParameterValueLength;
            _emptyParameterValue = emptyParameterValue;
        }

        #region Functions
        private static IEnumerable<string> Union(IEnumerable<string> list1, IEnumerable<string> list2)
        {
            IEnumerable<string> response;

            if (list1 == null && list2 == null)
            {
                response = null;
            }
            else if (list1 == null)
            {
                response = list2.Distinct();
            }
            else if (list2 == null)
            {
                response = list1.Distinct();
            }
            else
            {
                response = list1.Union(list2);
            }

            return response;
        }

        private static bool IsMatch(int? field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(Convert.ToString(field.Value), expression, options);
        }

        private static bool IsMatch(string field, string expression, RegexOptions options = RegexOptions.None)
        {
            if (field == null)
                return false;
            else
                return Regex.IsMatch(field, expression, options);
        }

        // This is just context and tuple counts where they are integer values - easier that changing the parsing logic to just return the value
        private static int Count(int count)
        {
            return count;
        }


        private static int Count<T>(IEnumerable<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static int Count<T>(ICollection<T> values)
        {
            return values == null ? 0 : values.Where(f => f != null).Count();
        }


        private static bool exists(bool value)
        {
            return value;
        }

        private static string GetValueOrEmpty(bool? val)
        {
            return (val.HasValue) ? val.ToString().ToLower() : string.Empty;
        }

        private static string GetValueOrEmptyDate(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-dd") : string.Empty;
        }

        private static string GetValueOrEmptyDate(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-dd");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(string val)
        {
            DateTime dateTimeOut;
            if (DateTime.TryParse(val, out dateTimeOut))
            {
                return dateTimeOut.ToString("yyyy-MM-ddTHH:mm:ssZ");
            }
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmptyDateTime(DateTime? val)
        {
            return (val.HasValue) ? val.Value.ToString("yyyy-MM-ddTHH:mm:ssZ") : string.Empty;
        }

        private static string GetValueOrEmpty(string val)
        {
            return !string.IsNullOrWhiteSpace(val) ? val : string.Empty;
        }

        private static string GetValueOrEmpty(decimal? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

        private static string GetValueOrEmpty(int? val)
        {
            return (val.HasValue) ? val.ToString() : string.Empty;
        }

         /// <summary>
         /// Get string value between [first] a and [last] b.
         /// </summary>
        public static string Between(string value, string a, string b)
        {
            int posA = value.IndexOf(a);
            int posB = value.LastIndexOf(b);
            if (posA == -1)
            {
                return "";
            }
            if (posB == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= posB)
            {
                return "";
            }
            return value.Substring(adjustedPosA, posB - adjustedPosA);
        }

         /// <summary>
         /// Get string value after [first] a.
         /// </summary>
        public static string Before(string value, string a)
        {
            int posA = value.IndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            return value.Substring(0, posA);
        }

         /// <summary>
         /// Get string value after [last] a.
         /// </summary>
        public static string After(string value, string a)
        {
            int posA = value.LastIndexOf(a);
            if (posA == -1)
            {
                return "";
            }
            int adjustedPosA = posA + a.Length;
            if (adjustedPosA >= value.Length)
            {
                return "";
            }
            return value.Substring(adjustedPosA);
        }

        private static int Length(object field)
        {
            if (field == null)
            return 0;
            else
            return field.ToString().Trim().Length;
        }

        private static bool NotSameValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(params object[] nodes)
        {
            if (nodes == null)
                return false;

            return (from x in nodes select x).Distinct().Count() != 1;
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool NotSameValues(IEnumerable<DateTime?> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return NotSameValues(nodesArray);
        }

        private static bool HasDuplicateValues(IEnumerable<object> nodes)
        {
            if (nodes == null)
                return false;

            object[] nodesArray = nodes.Cast<object>().ToArray();
            return HasDuplicateValues(nodesArray);
        }

        private static bool HasDuplicateValues(params object[] nodes)
        {
           if (nodes == null)
                return false;

            nodes = nodes.Where(x => x != null).ToArray();
            return !((from x in nodes select x).Distinct().Count() == nodes.Count());
        
        }

        private int DuplicateValueIndex(IEnumerable<object> values)
        {
            int response = 0;
            var hashset = new HashSet<object>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private int DuplicateValueIndex<T>(IEnumerable<T?> values) where T : struct
        {
            int response = 0;
            var hashset = new HashSet<T?>();
            foreach (var value in values)
            {
                if (!hashset.Add(value))
                {
                    return response;
                }
                response++;
            }
            return response;
        }

        private static bool IsDate(object value)
        {
            DateTime dateValue;
            return (value != null && DateTime.TryParse(value.ToString(), out dateValue));
        }

        public DateTime AsDate(string dateAsString)
        {
            DateTime response = DateTime.MinValue;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date;
            }

            return response;
        }

        public DateTime Date(DateTime? datetime)
        {                   
            return datetime.GetValueOrDefault().Date;
        }

				//The logic in After function expects "---" for day and "--" for month. 
				//Since hyphen was missing the date always returned null
        public DateTime? ConvertToDate(int day, int month, int year)
        {
					return ConvertToDate(day == 0 ? null : "---" + day.ToString(), month == 0 ? null : "--" + month.ToString(), year == 0 ? null : year.ToString());
        }

        public DateTime? ConvertToDate(string day, string month, string year)
        {
            DateTime? response;
            DateTime result;

            if (year == null || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        public DateTime? ConvertToDate(string day, string month, int year)
        {
            DateTime? response;
            DateTime result;

            if (year == 0 || month == null || day == null)
            {
                return null;
            }
            string dateAsString = year.ToString() + "-" + After(month, "--") + "-" + After(day, "---");
            if (DateTime.TryParse(dateAsString, out result))
            {
                response = result;
            }
            else
            {
                response = null;
            }
            return response;
        }


        private static int Day(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Day;
            }

            return response;
        }


        private static int Day(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Day;

        }


        private static string Month(string dateAsString)
        {
            string response = null;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.ToString("MMMM");
            }
            else
            {
                return "NotAMonth";
            }

            return response;
        }

        private static string Month(DateTime? date)
        {
            if (date == null)
                return "NotAMonth";
            else
                return date.Value.ToString("MMMM");
        }


        private static int MonthAsInt(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Month;
            }

            return response;
        }

        private static int MonthAsInt(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            return date.Value.Month;
        }


        private static int Year(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
            {
                response = date.Year;
            }

            return response;
        }

        private static int Year(DateTime? date)
        {
            if (date == null)
                return 0;
            else
                return date.Value.Year;
        }


        private static int CurrentFinancialYear()
        {
            return DateToFinancialYear(DateTime.Now, 7);
        }

        private static int FinancialYear(string dateAsString)
        {
            return DateToFinancialYear(dateAsString, 7);
        }

        private static int FinancialYear(DateTime? date)
        {
            return DateToFinancialYear(date, 7);
        }

        private static int DateToFinancialYear(string dateAsString, int startingMonth)
        {
            int response = 0;
            DateTime date;
            if (DateTime.TryParse(dateAsString, out date))
            {
                response = DateToFinancialYear(date, startingMonth);
            }

            return response;
        }
        private static int DateToFinancialYear(DateTime? date, int startingMonth)
        {
            int response;
            if (date == null)
            {
                response = 0;
            }
            else
            {
                int year = date.Value.Year;
                int month = date.Value.Month;

                if (startingMonth > month)
                    response = year;
                else
                    response = year + 1;
            }
            return response;
        }


        private static int FBTYear(string dateAsString)
        {
            int response = 0;
            DateTime date;

            if (DateTime.TryParse(dateAsString, out date))
                response = FBTYear(date);
            return response;
        }

        private static int FBTYear(DateTime? date)
        {
            if (date == null)
            {
                return 0;
            }
            else
            {
                if (date.Value.Month > 3)
                    return date.Value.Year + 1;
                else
                    return date.Value.Year;
            }
        }

        private static DateTime? AddMonthsToDate(DateTime? dateTime, int months)
        {
            return dateTime == null ? null : (DateTime?)dateTime.Value.AddMonths(months);
        }

        private static bool IsNumeric(object value)
        {
            decimal numbervalue;
            return (value != null && decimal.TryParse(value.ToString(), out numbervalue));
        }

        private static bool NotMonetary(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(decimal? field, string sign, int digits, int decimals)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;
                string decimalExpression;
                int digitsToUse = digits - decimals;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                if (decimals > 0)
                    decimalExpression = @"(\.\d{1," + decimals + "})?$";
                else
                    decimalExpression = @"$";

                return !(Regex.IsMatch(field.Value.ToString("0.#########################"), signExpression + @"\d{1," + digitsToUse + "}" + decimalExpression));
            }
        }

        private static bool NotNumeric(int? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }

        private static bool NotNumeric(long? field, string sign, int digits, int decimals = 0)
        {
            if (field == null)
            {
                return false;
            }
            else
            {
                string signExpression;

                if (sign == "U")
                    signExpression = "^";
                else
                    signExpression = "^-?";

                return !(Regex.IsMatch(field.Value.ToString(), signExpression + @"\d{1," + digits + "}$"));
            }
        }


        private static bool OutsideRange(decimal field, decimal expression, int range)
        {
            bool response;

            response = (field < (expression - range)) || (field > (expression + range));

            return response;
        }


        private static bool FailsUSIAlgorithm(string usi, string abn)
        {
            bool response = true;
            if (usi == null || abn == null)
            {
                response = false;
            }
            else
            {
                usi = usi.Trim();
                abn = abn.Trim();
                if (usi.Length == 14)
                {
                    int numeric;
                    if (usi.Substring(0, 11) == abn && int.TryParse(usi.Substring(11, 3), out numeric))
                        response = false;

                }
                else if (usi.Length == 9)
                {

                    if (Regex.IsMatch(usi, @"^([a-zA-Z]{3}\d{4}[a-zA-Z]{2})$"))
                        response = false;
                }
            }
            return response;
        }

        private static bool FailsTANAlgorithm(string tan)
        {
            bool response;
            decimal decimalTan;

            if (tan == null)
                return false;

            tan = tan.Trim();

            if (!decimal.TryParse(tan, out decimalTan))
                return true;

            if (tan.Length != 8)
                return true;

            decimal tanSum =
                7 * int.Parse(tan.Substring(0, 1)) +
                9 * int.Parse(tan.Substring(1, 1)) +
                8 * int.Parse(tan.Substring(2, 1)) +
                4 * int.Parse(tan.Substring(3, 1)) +
                6 * int.Parse(tan.Substring(4, 1)) +
                3 * int.Parse(tan.Substring(5, 1)) +
                5 * int.Parse(tan.Substring(6, 1)) +
                1 * int.Parse(tan.Substring(7, 1));

            if ((tanSum % 11) == 0)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsABNAlgorithm(string abn)
        {
            bool response;
            decimal decimalAbn;

            if (abn == null)
                return false;

            abn = abn.Trim();

            if (!decimal.TryParse(abn, out decimalAbn))
                return true;

            if (abn.Length != 11)
                return true;

            decimal abnSum =
                10 * (int.Parse(abn.Substring(0, 1)) - 1) +
                1 * int.Parse(abn.Substring(1, 1)) +
                3 * int.Parse(abn.Substring(2, 1)) +
                5 * int.Parse(abn.Substring(3, 1)) +
                7 * int.Parse(abn.Substring(4, 1)) +
                9 * int.Parse(abn.Substring(5, 1)) +
                11 * int.Parse(abn.Substring(6, 1)) +
                13 * int.Parse(abn.Substring(7, 1)) +
                15 * int.Parse(abn.Substring(8, 1)) +
                17 * int.Parse(abn.Substring(9, 1)) +
                19 * int.Parse(abn.Substring(10, 1));

            if ((abnSum % 89) == 0)
                response = false;
            else
                response = true;

            return response;
        }
        private static bool FailsACNAlgorithm(string acn)
        {
            bool response;
            decimal decimalAbn;
            if (acn == null)
                return false;

            acn = acn.Trim();

            if (!decimal.TryParse(acn, out decimalAbn))
                return true;

            if (acn.Length != 9)
                return true;

            decimal abnSum =
                8 * int.Parse(acn.Substring(0, 1)) +
                7 * int.Parse(acn.Substring(1, 1)) +
                6 * int.Parse(acn.Substring(2, 1)) +
                5 * int.Parse(acn.Substring(3, 1)) +
                4 * int.Parse(acn.Substring(4, 1)) +
                3 * int.Parse(acn.Substring(5, 1)) +
                2 * int.Parse(acn.Substring(6, 1)) +
                1 * int.Parse(acn.Substring(7, 1));

            decimal checkDigit = int.Parse(acn.Substring(8, 1));
            decimal acnRemainder = abnSum % 10;

            if (((10 - acnRemainder) % 10) == checkDigit)
                response = false;
            else
                response = true;

            return response;
        }


        private static bool FailsTFNAlgorithm(string tfn)
        {
            bool response;
            decimal decimalTfn;

            if (tfn == null)
                return false;

            tfn = tfn.Trim();
            tfn = Regex.Replace(tfn, "^0+", "");

            if (!decimal.TryParse(tfn, out decimalTfn))
                return true;

            if (tfn.Length < 8)
                return true;


            decimal tfn1To7Sum =
                1 * int.Parse(tfn.Substring(0, 1)) +
                4 * int.Parse(tfn.Substring(1, 1)) +
                3 * int.Parse(tfn.Substring(2, 1)) +
                7 * int.Parse(tfn.Substring(3, 1)) +
                5 * int.Parse(tfn.Substring(4, 1)) +
                8 * int.Parse(tfn.Substring(5, 1)) +
                6 * int.Parse(tfn.Substring(6, 1));

            decimal tfn8 = 9 * int.Parse(tfn.Substring(7, 1));

            if (tfn.Length == 8)
            {
                decimal tFNLg8WSum9 = 10 * int.Parse(tfn.Substring(7, 1));
                decimal tFNLg8WSum = tfn1To7Sum + tFNLg8WSum9;

                if ((tFNLg8WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else if (tfn.Length == 9)
            {
                decimal tfn9 = 10 * int.Parse(tfn.Substring(8, 1));
                decimal tFNLg9WSum = tfn1To7Sum + tfn8 + tfn9;

                if ((tFNLg9WSum % 11) == 0)
                    response = false;
                else
                    response = true;
            }
            else
            {
                response = true;
            }

            return response;
        }


        private static decimal ConditionalValue(bool expression, decimal? trueVal, decimal? falseVal)
        {
             return expression ? trueVal.GetValueOrDefault() : falseVal.GetValueOrDefault();
        }

        private static decimal AsNumeric(string value)
        {
             decimal numberValue;
             decimal.TryParse(value, out numberValue);
             return numberValue;
        }

        private static bool RegexMatch(int? field, string expression, string flags="")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static bool RegexMatch(string field, string expression, string flags="")
        {
            return IsMatch(field, expression, GetRegexOption(flags));
        }

        private static RegexOptions GetRegexOption(string flags)
        {
            RegexOptions options = RegexOptions.None;

            char[] characters = flags.ToCharArray();

            foreach (char character in characters)
            {
                switch (character)
                {
                    case 'i':                        
                        options = options | RegexOptions.IgnoreCase;                       
                        break;
                    case 'm':                        
                        options = options | RegexOptions.Multiline;                       
                        break;
                    case 's':                        
                            options = options | RegexOptions.Singleline;                       
                        break;
                    case 'n':                        
                            options = options | RegexOptions.ExplicitCapture;                       
                        break;
                    case 'x':                       
                            options = options | RegexOptions.IgnorePatternWhitespace;                       
                        break;
                }
            }

            return options;
        }

        private static string Substring(object field, int start, int length)
        {
            return field.ToString().Substring(start, length);
        }
        
        /// <summary>
        /// Returns an occurrence index as [occurrenceIndex] of occurrenceIndex > 0, otherwise the empty string
        /// </summary>
        /// <param name="occurrenceIndex">Index of the occurrence.</param>
        /// <returns>Occurrence in XPath [#] format</returns>
        public string OccurrenceIndex(int occurrenceIndex)
        {
            return occurrenceIndex > 0 ? "[" + occurrenceIndex + "]" : "";
        }

        #endregion // Functions

        public TPARPYE2021 ConsumedReport { get; private set; }

        private static Dictionary<string, ProcessMessageDocument> _processMessageDocuments = new Dictionary<string,ProcessMessageDocument>();

        [MethodImpl(MethodImplOptions.AggressiveInlining)]
        public List<ProcessMessageDocument> ValidateReport(TPARPYE2021 report)
        {

            List<ProcessMessageDocument> response = new List<ProcessMessageDocument>();
            ProcessMessageDocument processMessage;
            ProcessMessageParameter parameter;
            bool assertion;

            this.ConsumedReport = report;

            // ------------------------------------------------------------------------------
            // Validations are now done in code.
            // This version supports full validation of a report - for those rules that have successfully parsed and could be generated.
            // Generated Validations include:
            //   - Validation logic in C#.
            //   - Production of errors in EBMS, SBR and SWS format.
            // 
            // The generation at this stage does not support the following - and has to be completed manually (later versions of the generation will do this for you)
            //   - Business Rules that did not parse.  The ESR team will help support you where this happens as an effort is being made to rectify these
            //   - You will get TODO tasks for those that the parser could not cope with
            // ------------------------------------------------------------------------------

    
            #region VR.ATO.TPARPYE.000001
    
            /*  VR.ATO.TPARPYE.000001
            If the payment type is "Grant", then Date of payment must be supplied
    
            Legacy Rule Format:
            (^TPARPYE137 = "Grant" AND ^TPARPYE135 = NULL)

            Technical Business Rule Format:
            ^TPARPYE137 = 'Grant' AND ^TPARPYE135 = NULL
    
            Data Elements:
    
            ^TPARPYE135 = TPARPYE:Payee:GovernmentFunding.GrantPayment.Date
    
            ^TPARPYE137 = TPARPYE:Payee:IncomeTax.PaymentType.Code
            */
            assertion = (report.TPARPYE137 == @"Grant" && report.TPARPYE135 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.000001", Severity = ProcessMessageSeverity.Error,
                    Description = @"If the payment type is ""Grant"", then Date of payment must be supplied",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingGrantPaymentD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.000001"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE137", Value = report.TPARPYE137 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE135", Value = GetValueOrEmptyDate(report.TPARPYE135) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.000001
    
            #region VR.ATO.TPARPYE.000002
    
            /*  VR.ATO.TPARPYE.000002
            If Payment type field is "Payment", then this field should be blank.
    
            Legacy Rule Format:
            (^TPARPYE137 = "Payment" AND ^TPARPYE138 <> NULL)

            Technical Business Rule Format:
            ^TPARPYE137 = 'Payment' AND ^TPARPYE138 <> NULL
    
            Data Elements:
    
            ^TPARPYE138 = TPARPYE:Payee:GovernmentFunding.ITAADivision59Indicator.Code
    
            ^TPARPYE137 = TPARPYE:Payee:IncomeTax.PaymentType.Code
            */
            assertion = (report.TPARPYE137 == @"Payment" && report.TPARPYE138 != null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.000002", Severity = ProcessMessageSeverity.Error,
                    Description = @"If Payment type field is ""Payment"", then this field should be blank.",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingITAADivision59IndicatorC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.000002"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE137", Value = report.TPARPYE137 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE138", Value = GetValueOrEmpty(report.TPARPYE138) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.000002
    
            #region VR.ATO.TPARPYE.W00001
    
            /*  VR.ATO.TPARPYE.W00001
            Overseas address indicator must not be 'TRUE' if Country Code is 'au'
    
            Legacy Rule Format:
            (^TPARPYE114 = "au" AND  ^TPARPYE103 = TRUE)

            Technical Business Rule Format:
            ^TPARPYE114 = 'au' AND  ^TPARPYE103 = TRUE
    
            Data Elements:
    
            ^TPARPYE103 = TPARPYE:Payee:AddressDetails:AddressDetails.OverseasAddress.Indicator
    
            ^TPARPYE114 = TPARPYE:Payee:AddressDetails:AddressDetails.Country.Code
            */
            assertion = (report.TPARPYE114 == @"au" && report.TPARPYE103 == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00001", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Overseas address indicator must not be 'TRUE' if Country Code is 'au'",
                    Location = "/tns:TPARPYE/tns:Payee/tns:AddressDetails/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00001"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE114", Value = GetValueOrEmpty(report.TPARPYE114) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE103", Value = GetValueOrEmpty(report.TPARPYE103) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00001
    
            #region VR.ATO.TPARPYE.W00005
    
            /*  VR.ATO.TPARPYE.W00005
            If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'
    
            Legacy Rule Format:
            (^TPARPYE114 <> NULLORBLANK AND (^TPARPYE114 <> "au" AND  ^TPARPYE103 = FALSE))

            Technical Business Rule Format:
            (^TPARPYE114 <> BLANK AND (^TPARPYE114 <> 'au' AND  ^TPARPYE103 = FALSE))
    
            Data Elements:
    
            ^TPARPYE114 = TPARPYE:Payee:AddressDetails:AddressDetails.Country.Code
    
            ^TPARPYE103 = TPARPYE:Payee:AddressDetails:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE114) != true && (report.TPARPYE114 != @"au" && report.TPARPYE103 == false));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00005", Severity = ProcessMessageSeverity.Warning,
                    Description = @"If an overseas Country Code is used, the Overseas Address indicator must be 'TRUE'",
                    Location = "/tns:TPARPYE/tns:Payee/tns:AddressDetails/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00005"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE114", Value = GetValueOrEmpty(report.TPARPYE114) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE103", Value = GetValueOrEmpty(report.TPARPYE103) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00005
    
            #region VR.ATO.TPARPYE.W00016
    
            /*  VR.ATO.TPARPYE.W00016
            Payee ABN must be a valid ABN
    
            Legacy Rule Format:
            (^TPARPYE84 <> NULLORBLANK AND FailsABNAlgorithm(^TPARPYE84))

            Technical Business Rule Format:
            (^TPARPYE84 <> BLANK AND FailsABNAlgorithm(^TPARPYE84))
    
            Data Elements:
    
            ^TPARPYE84 = TPARPYE:Payee:Identifiers.AustralianBusinessNumber.Identifier
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE84) != true && FailsABNAlgorithm(report.TPARPYE84));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00016", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Payee ABN not valid",
                    Location = "/tns:TPARPYE/tns:Payee/tns:AustralianBusinessNumberId",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00016"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE84", Value = GetValueOrEmpty(report.TPARPYE84) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00016
    
            #region VR.ATO.TPARPYE.W00019
    
            /*  VR.ATO.TPARPYE.W00019
            Payee family name must contain at least one alphabetical character
    
            Legacy Rule Format:
            (^TPARPYE91 <> NULLORBLANK AND NotContainsSet(^TPARPYE91, '"A-Z","a-z"'))

            Technical Business Rule Format:
            (^TPARPYE91 <> BLANK AND NotContainsSet(^TPARPYE91, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^TPARPYE91 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.FamilyName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE91) != true && !(IsMatch(report.TPARPYE91, @"[A-Z]|[a-z]",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00019", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Payee family name must contain at least one alpha character",
                    Location = "/tns:TPARPYE/tns:Payee/tns:PersonNameDetails/tns:FamilyNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00019"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE91", Value = report.TPARPYE91 });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00019
    
            #region VR.ATO.TPARPYE.W00020
    
            /*  VR.ATO.TPARPYE.W00020
            Payee given name must contain at least one alphabetical character
    
            Legacy Rule Format:
            (^TPARPYE92 <> NULLORBLANK AND NotContainsSet(^TPARPYE92, '"A-Z","a-z"'))

            Technical Business Rule Format:
            (^TPARPYE92 <> BLANK AND NotContainsSet(^TPARPYE92, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^TPARPYE92 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.GivenName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE92) != true && !(IsMatch(report.TPARPYE92, @"[A-Z]|[a-z]",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00020", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Payee given name must contain at least one alpha character",
                    Location = "/tns:TPARPYE/tns:Payee/tns:PersonNameDetails/tns:GivenNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00020"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE92", Value = GetValueOrEmpty(report.TPARPYE92) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00020
    
            #region VR.ATO.TPARPYE.W00021
    
            /*  VR.ATO.TPARPYE.W00021
            Payee other given name must contain at least one alphabetical character
    
            Legacy Rule Format:
            (^TPARPYE93 <> NULLORBLANK AND NotContainsSet(^TPARPYE93, '"A-Z","a-z"'))

            Technical Business Rule Format:
            (^TPARPYE93 <> BLANK AND NotContainsSet(^TPARPYE93, '"A-Z","a-z"'))
    
            Data Elements:
    
            ^TPARPYE93 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.OtherGivenName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE93) != true && !(IsMatch(report.TPARPYE93, @"[A-Z]|[a-z]",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00021", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Payee other given name must contain at least one alpha character",
                    Location = "/tns:TPARPYE/tns:Payee/tns:PersonNameDetails/tns:OtherGivenNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00021"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE93", Value = GetValueOrEmpty(report.TPARPYE93) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00021
    
            #region VR.ATO.TPARPYE.W00022
    
            /*  VR.ATO.TPARPYE.W00022
            Payee individual or non-individual name should be present.
    
            Legacy Rule Format:
            (^TPARPYE91 = NULLORBLANK AND ^TPARPYE92 = NULLORBLANK AND ^TPARPYE93 = NULLORBLANK AND ^TPARPYE97 = NULLORBLANK AND ^TPARPYE101 = NULLORBLANK)

            Technical Business Rule Format:
            ^TPARPYE91 = BLANK AND ^TPARPYE92 = BLANK AND ^TPARPYE93 = BLANK AND ^TPARPYE97 = BLANK AND ^TPARPYE101 = BLANK
    
            Data Elements:
    
            ^TPARPYE97 = TPARPYE:Payee:OrganisationNameDetailsMain:OrganisationNameDetails.OrganisationalName.Text
    
            ^TPARPYE91 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.FamilyName.Text
    
            ^TPARPYE92 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.GivenName.Text
    
            ^TPARPYE93 = TPARPYE:Payee:PersonNameDetails:PersonNameDetails.OtherGivenName.Text
    
            ^TPARPYE101 = TPARPYE:Payee:OrganisationNameDetailsTrading:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE91) == true && string.IsNullOrWhiteSpace(report.TPARPYE92) == true && string.IsNullOrWhiteSpace(report.TPARPYE93) == true && string.IsNullOrWhiteSpace(report.TPARPYE97) == true && string.IsNullOrWhiteSpace(report.TPARPYE101) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00022", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Payee individual or non-individual name should be supplied",
                    Location = "/tns:TPARPYE/tns:Payee/tns:OrganisationNameDetailsMainOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00022"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE91", Value = report.TPARPYE91 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE92", Value = GetValueOrEmpty(report.TPARPYE92) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE93", Value = GetValueOrEmpty(report.TPARPYE93) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE97", Value = report.TPARPYE97 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE101", Value = report.TPARPYE101 });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00022
    
            #region VR.ATO.TPARPYE.W00023
    
            /*  VR.ATO.TPARPYE.W00023
            Organisation name must contain at least one alpha character
    
            Legacy Rule Format:
            (^TPARPYE97 <> NULLORBLANK AND NotContainsSet(^TPARPYE97, '"a-z", "A-Z"'))

            Technical Business Rule Format:
            (^TPARPYE97 <> BLANK AND NotContainsSet(^TPARPYE97, '"a-z", "A-Z"'))
    
            Data Elements:
    
            ^TPARPYE97 = TPARPYE:Payee:OrganisationNameDetailsMain:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE97) != true && !(IsMatch(report.TPARPYE97, @"[a-z]|[A-Z]",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00006", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Organisation name must contain at least one alpha character",
                    Location = "/tns:TPARPYE/tns:Payee/tns:OrganisationNameDetailsMainOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00023"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE97", Value = report.TPARPYE97 });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00023
    
            #region VR.ATO.TPARPYE.W00024
    
            /*  VR.ATO.TPARPYE.W00024
            Organisation name must contain at least one alpha character
    
            Legacy Rule Format:
            (^TPARPYE101 <> NULLORBLANK AND NotContainsSet(^TPARPYE101, '"a-z", "A-Z"'))

            Technical Business Rule Format:
            (^TPARPYE101 <> BLANK AND NotContainsSet(^TPARPYE101, '"a-z", "A-Z"'))
    
            Data Elements:
    
            ^TPARPYE101 = TPARPYE:Payee:OrganisationNameDetailsTrading:OrganisationNameDetails.OrganisationalName.Text
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE101) != true && !(IsMatch(report.TPARPYE101, @"[a-z]|[A-Z]",RegexOptions.IgnoreCase)));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00006", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Organisation name must contain at least one alpha character",
                    Location = "/tns:TPARPYE/tns:Payee/tns:OrganisationNameDetailsTradingOrganisationalNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00024"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE101", Value = report.TPARPYE101 });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00024
    
            #region VR.ATO.TPARPYE.W00025
    
            /*  VR.ATO.TPARPYE.W00025
            Payee address state and post code must be present for an Australian address
    
            Legacy Rule Format:
            (^TPARPYE103= FALSE AND (^TPARPYE111 = NULL OR ^TPARPYE112 = NULL))

            Technical Business Rule Format:
            (^TPARPYE103= FALSE AND (^TPARPYE111 = NULL OR ^TPARPYE112 = NULL))
    
            Data Elements:
    
            ^TPARPYE103 = TPARPYE:Payee:AddressDetails:AddressDetails.OverseasAddress.Indicator
    
            ^TPARPYE111 = TPARPYE:Payee:AddressDetails:AddressDetails.Postcode.Text
    
            ^TPARPYE112 = TPARPYE:Payee:AddressDetails:AddressDetails.StateOrTerritory.Code
            */
            assertion = (report.TPARPYE103 == false && (report.TPARPYE111 == null || report.TPARPYE112 == null));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00008", Severity = ProcessMessageSeverity.Warning,
                    Description = @"State and post code not supplied for an Australian address",
                    Location = "/tns:TPARPYE/tns:Payee/tns:AddressDetails/tns:OverseasAddressI",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00025"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE103", Value = GetValueOrEmpty(report.TPARPYE103) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE111", Value = GetValueOrEmpty(report.TPARPYE111) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE112", Value = GetValueOrEmpty(report.TPARPYE112) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00025
    
            #region VR.ATO.TPARPYE.W00028
    
            /*  VR.ATO.TPARPYE.W00028
            Country code must be present for overseas address
    
            Legacy Rule Format:
            (^TPARPYE103 = TRUE AND ^TPARPYE114 = NULL)

            Technical Business Rule Format:
            ^TPARPYE103 = TRUE AND ^TPARPYE114 = NULL
    
            Data Elements:
    
            ^TPARPYE114 = TPARPYE:Payee:AddressDetails:AddressDetails.Country.Code
    
            ^TPARPYE103 = TPARPYE:Payee:AddressDetails:AddressDetails.OverseasAddress.Indicator
            */
            assertion = (report.TPARPYE103 == true && report.TPARPYE114 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPAR.W00011", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Country code must be present for overseas address",
                    Location = "/tns:TPARPYE/tns:Payee/tns:AddressDetails/tns:CountryC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00028"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE103", Value = GetValueOrEmpty(report.TPARPYE103) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE114", Value = GetValueOrEmpty(report.TPARPYE114) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00028
    
            #region VR.ATO.TPARPYE.W00029
    
            /*  VR.ATO.TPARPYE.W00029
            BSB Number must be six digits and greater than 012000
    
            Legacy Rule Format:
            (^TPARPYE120 <> NULLORBLANK AND (^TPARPYE120 < 012001 OR ^TPARPYE120 > 999999 OR LENGTH(^TPARPYE120) <> 6))

            Technical Business Rule Format:
            (^TPARPYE120 <> BLANK AND (^TPARPYE120 < 012001 OR ^TPARPYE120 > 999999 OR LENGTH(^TPARPYE120) <> 6))
    
            Data Elements:
    
            ^TPARPYE120 = TPARPYE:Payee:FinancialInstitutionAccount.BankStateBranch.Number
            */
            assertion = (string.IsNullOrWhiteSpace(report.TPARPYE120) != true && (AsNumeric(report.TPARPYE120) < 12001 || AsNumeric(report.TPARPYE120) > 999999 || Length(report.TPARPYE120) != 6));
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00029", Severity = ProcessMessageSeverity.Warning,
                    Description = @"BSB Number must be six digits and greater than 012000",
                    Location = "/tns:TPARPYE/tns:Payee/tns:FinancialInstitutionAccountBankStateBranchN",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00029"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE120", Value = GetValueOrEmpty(report.TPARPYE120) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00029
    
            #region VR.ATO.TPARPYE.W00030
    
            /*  VR.ATO.TPARPYE.W00030
            Gross Amount Paid should be present
    
            Legacy Rule Format:
            (^TPARPYE122 = NULL)

            Technical Business Rule Format:
            ^TPARPYE122 = NULL
    
            Data Elements:
    
            ^TPARPYE122 = TPARPYE:Payee:Income.BusinessPaymentGross.Amount
            */
            assertion = (report.TPARPYE122 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00030", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Gross Amount Paid not supplied",
                    Location = "/tns:TPARPYE/tns:Payee/tns:IncomeBusinessPaymentGrossA",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00030"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE122", Value = GetValueOrEmpty(report.TPARPYE122) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00030
    
            #region VR.ATO.TPARPYE.W00031
    
            /*  VR.ATO.TPARPYE.W00031
            Total Tax Withheld Where ABN Not Quoted amount must not be greater than Gross Amount Paid
    
            Legacy Rule Format:
            (^TPARPYE123 <> NULL AND ^TPARPYE123 > ^TPARPYE122)

            Technical Business Rule Format:
            ^TPARPYE123 <> NULL AND ^TPARPYE123 > ^TPARPYE122
    
            Data Elements:
    
            ^TPARPYE123 = TPARPYE:Payee:IncomeTax.PayAsYouGoWithholding.TaxWithheld.Amount
    
            ^TPARPYE122 = TPARPYE:Payee:Income.BusinessPaymentGross.Amount
            */
            assertion = (report.TPARPYE123 != null && report.TPARPYE123.GetValueOrDefault() > report.TPARPYE122.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00031", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Total Tax Withheld Where ABN Not Quoted amount greater than Gross Amount Paid",
                    Location = "/tns:TPARPYE/tns:Payee/tns:IncomeTaxPayAsYouGoWithholdingTaxWithheldA",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00031"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE123", Value = GetValueOrEmpty(report.TPARPYE123) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE122", Value = GetValueOrEmpty(report.TPARPYE122) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00031
    
            #region VR.ATO.TPARPYE.W00032
    
            /*  VR.ATO.TPARPYE.W00032
            Total GST amount must not be greater than Gross Amount Paid
    
            Legacy Rule Format:
            (^TPARPYE124 <> NULL AND ^TPARPYE124 > ^TPARPYE122)

            Technical Business Rule Format:
            ^TPARPYE124 <> NULL AND ^TPARPYE124 > ^TPARPYE122
    
            Data Elements:
    
            ^TPARPYE124 = TPARPYE:Payee:GoodsAndServicesTax.Liability.Amount
    
            ^TPARPYE122 = TPARPYE:Payee:Income.BusinessPaymentGross.Amount
            */
            assertion = (report.TPARPYE124 != null && report.TPARPYE124.GetValueOrDefault() > report.TPARPYE122.GetValueOrDefault());
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00032", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Total GST amount greater than Gross Amount Paid",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GoodsAndServicesTaxLiabilityA",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00032"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE124", Value = GetValueOrEmpty(report.TPARPYE124) });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE122", Value = GetValueOrEmpty(report.TPARPYE122) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00032
    
            #region VR.ATO.TPARPYE.W00034
    
            /*  VR.ATO.TPARPYE.W00034
            If the payment type is "Grant", then Name of grant/program must be supplied
    
            Legacy Rule Format:
            (^TPARPYE137 = "Grant" AND ^TPARPYE134 = NULLORBLANK)

            Technical Business Rule Format:
            ^TPARPYE137 = 'Grant' AND ^TPARPYE134 = BLANK
    
            Data Elements:
    
            ^TPARPYE134 = TPARPYE:Payee:GovernmentFunding.GrantProgramName.Text
    
            ^TPARPYE137 = TPARPYE:Payee:IncomeTax.PaymentType.Code
            */
            assertion = (report.TPARPYE137 == @"Grant" && string.IsNullOrWhiteSpace(report.TPARPYE134) == true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00034", Severity = ProcessMessageSeverity.Warning,
                    Description = @"If the payment type is ""Grant"", then Name of grant/program must be supplied",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingGrantProgramNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00034"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE137", Value = report.TPARPYE137 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE134", Value = GetValueOrEmpty(report.TPARPYE134) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00034
    
            #region VR.ATO.TPARPYE.W00035
    
            /*  VR.ATO.TPARPYE.W00035
            If the payment type is "Payment", then Name of grant/program must not be supplied
    
            Legacy Rule Format:
            (^TPARPYE137 = "Payment" AND ^TPARPYE134 <> NULLORBLANK)

            Technical Business Rule Format:
            ^TPARPYE137 = 'Payment' AND ^TPARPYE134 <> BLANK
    
            Data Elements:
    
            ^TPARPYE134 = TPARPYE:Payee:GovernmentFunding.GrantProgramName.Text
    
            ^TPARPYE137 = TPARPYE:Payee:IncomeTax.PaymentType.Code
            */
            assertion = (report.TPARPYE137 == @"Payment" && string.IsNullOrWhiteSpace(report.TPARPYE134) != true);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00035", Severity = ProcessMessageSeverity.Warning,
                    Description = @"If the payment type is ""Payment"", then Name of grant/program must not be supplied",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingGrantProgramNameT",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00035"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE137", Value = report.TPARPYE137 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE134", Value = GetValueOrEmpty(report.TPARPYE134) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00035
    
            #region VR.ATO.TPARPYE.W00036
    
            /*  VR.ATO.TPARPYE.W00036
            Date of payment must be the same year as the financial year the payment was made
    
            Legacy Rule Format:
            (^TPARPYE135 <> NULL AND
            (FinancialYear(^TPARPYE135) <> ^TPAR47))

            Technical Business Rule Format:
            (^TPARPYE135 <> NULL AND
            (FinancialYear(^TPARPYE135) <> ^TPAR47))
    
            Data Elements:
    
            ^TPARPYE135 = TPARPYE:Payee:GovernmentFunding.GrantPayment.Date
    
            ^TPAR47 = TPAR47
            */
            assertion = false; // = (report.TPARPYE135 != null && FinancialYear(report.TPARPYE135.GetValueOrDefault()) != );
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00036", Severity = ProcessMessageSeverity.Warning,
                    Description = @"Date of payment must be the same year as the financial year the payment was made",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingGrantPaymentD",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00036"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE135", Value = GetValueOrEmptyDate(report.TPARPYE135) });
    
                response.Add(processMessage);
            }
                /* 
                TODO: Fix Transform Error:  Could not find Business Element in rule VR.ATO.TPARPYE.W00036 for TPAR47
                            */
                            // ------------------------------------------------------------------------------
                        #endregion // VR.ATO.TPARPYE.W00036
    
            #region VR.ATO.TPARPYE.W00037
    
            /*  VR.ATO.TPARPYE.W00037
            If Payment type field is "Grant", then this field should contain the values N, Y or U.
    
            Legacy Rule Format:
            (^TPARPYE137 = "Grant" AND ^TPARPYE138 = NULL)

            Technical Business Rule Format:
            ^TPARPYE137 = 'Grant' AND ^TPARPYE138 = NULL
    
            Data Elements:
    
            ^TPARPYE138 = TPARPYE:Payee:GovernmentFunding.ITAADivision59Indicator.Code
    
            ^TPARPYE137 = TPARPYE:Payee:IncomeTax.PaymentType.Code
            */
            assertion = (report.TPARPYE137 == @"Grant" && report.TPARPYE138 == null);
            if (assertion)
            {
                processMessage = new ProcessMessageDocument()
                {
                    Code = "CMN.ATO.TPARPYE.W00037", Severity = ProcessMessageSeverity.Warning,
                    Description = @"If Payment type field is ""Grant"", then this field should contain the values N, Y or U.",
                    Location = "/tns:TPARPYE/tns:Payee/tns:GovernmentFundingITAADivision59IndicatorC",
                    Parameters = new ProcessMessageParameters() { new ProcessMessageParameter() {Name = "RuleIdentifier", Value = "VR.ATO.TPARPYE.W00037"} },
                };
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE137", Value = report.TPARPYE137 });
    
                processMessage.Parameters.Add(new ProcessMessageParameter
                    { Name = "TPARPYE138", Value = GetValueOrEmpty(report.TPARPYE138) });
    
                response.Add(processMessage);
            }
            #endregion // VR.ATO.TPARPYE.W00037

            foreach (ProcessMessageDocument currentProcessMessage in response)
            {
                if (currentProcessMessage.Parameters != null)
                {
                    foreach (ProcessMessageParameter currentParameter in currentProcessMessage.Parameters)
                    {
                        if (string.IsNullOrEmpty(currentParameter.Name))
                        {
                            currentParameter.Name = _emptyParameterValue;
                        }

                        if (currentParameter.Name.Length > _maxParameterNameLength)
                        {
                            currentParameter.Name = currentParameter.Name.Substring(0, _maxParameterNameLength - 1);
                        }

                        if (string.IsNullOrEmpty(currentParameter.Value))
                        {
                            currentParameter.Value = _emptyParameterValue;
                        }

                        if (currentParameter.Value.Length > _maxParameterValueLength)
                        {
                            currentParameter.Value = currentParameter.Value.Substring(0, _maxParameterValueLength - 1);
                        }
                    }
                }
            }

            return response;
        }


    }
} 
